@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
  * USA
  */
 
@@ -70,12 +70,36 @@ pyg_type_wrapper_dealloc(PyGTypeWrapper *self)
     PyObject_DEL(self);
 }
 
+static GQuark
+_pyg_type_key(GType type) {
+    GQuark key;
+
+    if (g_type_is_a(type, G_TYPE_INTERFACE)) {
+        key = pyginterface_type_key;
+    } else if (g_type_is_a(type, G_TYPE_ENUM)) {
+        key = pygenum_class_key;
+    } else if (g_type_is_a(type, G_TYPE_FLAGS)) {
+        key = pygflags_class_key;
+    } else if (g_type_is_a(type, G_TYPE_POINTER)) {
+        key = pygpointer_class_key;
+    } else if (g_type_is_a(type, G_TYPE_BOXED)) {
+        key = pygboxed_type_key;
+    } else {
+        key = pygobject_class_key;
+    }
+
+    return key;
+}
+
 static PyObject *
 _wrap_g_type_wrapper__get_pytype(PyGTypeWrapper *self, void *closure)
 {
+    GQuark key;
     PyObject *py_type;
-    
-    py_type = g_type_get_qdata(self->type, pygobject_class_key);
+
+    key = _pyg_type_key(self->type);
+
+    py_type = g_type_get_qdata(self->type, key);
     if (!py_type)
       py_type = Py_None;
 
@@ -83,6 +107,29 @@ _wrap_g_type_wrapper__get_pytype(PyGTypeWrapper *self, void *closure)
     return py_type;
 }
 
+static int
+_wrap_g_type_wrapper__set_pytype(PyGTypeWrapper *self, PyObject* value, void *closure)
+{
+    GQuark key;
+    PyObject *py_type;
+
+    key = _pyg_type_key(self->type);
+
+    py_type = g_type_get_qdata(self->type, key);
+    Py_CLEAR(py_type);
+    if (value == Py_None)
+	g_type_set_qdata(self->type, key, NULL);
+    else if (PyType_Check(value)) {
+	Py_INCREF(value);
+	g_type_set_qdata(self->type, key, value);
+    } else {
+	PyErr_SetString(PyExc_TypeError, "Value must be None or a type object");
+	return -1;
+    }
+
+    return 0;
+}
+
 static PyObject *
 _wrap_g_type_wrapper__get_name(PyGTypeWrapper *self, void *closure)
 {
@@ -143,7 +190,7 @@ _wrap_g_type_wrapper__get_depth(PyGTypeWrapper *self, void *closure)
 }
 
 static PyGetSetDef _PyGTypeWrapper_getsets[] = {
-    { "pytype", (getter)_wrap_g_type_wrapper__get_pytype, (setter)0 },
+    { "pytype", (getter)_wrap_g_type_wrapper__get_pytype, (setter)_wrap_g_type_wrapper__set_pytype },
     { "name",  (getter)_wrap_g_type_wrapper__get_name, (setter)0 },
     { "fundamental",  (getter)_wrap_g_type_wrapper__get_fundamental, (setter)0 },
     { "parent",  (getter)_wrap_g_type_wrapper__get_parent, (setter)0 },
@@ -397,6 +444,16 @@ pyg_enum_get_value(GType enum_type, PyObject *obj, gint *val)
 	*val = _PyLong_AsLong(obj);
 	res = 0;
 
+	if (PyObject_TypeCheck(obj, &PyGEnum_Type) && ((PyGEnum *) obj)->gtype != enum_type) {
+	    g_warning("expected enumeration type %s, but got %s instead",
+		      g_type_name(enum_type),
+		      g_type_name(((PyGEnum *) obj)->gtype));
+	}
+    /* Dumb code duplication, but probably not worth it to have yet another macro. */
+    } else if (PyLong_Check(obj)) {
+	*val = PyLong_AsLong(obj);
+	res = 0;
+
 	if (PyObject_TypeCheck(obj, &PyGEnum_Type) && ((PyGEnum *) obj)->gtype != enum_type) {
 	    g_warning("expected enumeration type %s, but got %s instead",
 		      g_type_name(enum_type),
@@ -696,11 +753,6 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	break;
     case G_TYPE_INT:
 	g_value_set_int(value, _PyLong_AsLong(obj));
-	if (PyErr_Occurred()) {
-	    g_value_unset(value);
-	    PyErr_Clear();
-	    return -1;
-	}
 	break;
     case G_TYPE_UINT:
 	{
@@ -714,21 +766,11 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 		    return -1;
 	    } else {
 		g_value_set_uint(value, PyLong_AsUnsignedLong(obj));
-		if (PyErr_Occurred()) {
-		    g_value_unset(value);
-		    PyErr_Clear();
-		    return -1;
-		}
 	    }
 	}
 	break;
     case G_TYPE_LONG:
 	g_value_set_long(value, _PyLong_AsLong(obj));
-	if (PyErr_Occurred()) {
-	    g_value_unset(value);
-	    PyErr_Clear();
-	    return -1;
-	}
 	break;
     case G_TYPE_ULONG:
 	{
@@ -742,26 +784,24 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 		    return -1;
 	    } else {
 		g_value_set_ulong(value, PyLong_AsUnsignedLong(obj));
-		if (PyErr_Occurred()) {
-		    g_value_unset(value);
-		    PyErr_Clear();
-		    return -1;
-		}
 	    }
 	}
 	break;
     case G_TYPE_INT64:
 	g_value_set_int64(value, PyLong_AsLongLong(obj));
-	if (PyErr_Occurred()) {
-	    g_value_unset(value);
-	    PyErr_Clear();
-	    return -1;
-	}
 	break;
     case G_TYPE_UINT64:
-        if (_PyLong_Check(obj))
-            g_value_set_uint64(value, _PyLong_AsLong(obj));
-        else if (PyLong_Check(obj))
+#if PY_VERSION_HEX < 0x03000000
+        if (PyInt_Check(obj)) {
+            long v = PyInt_AsLong(obj);
+            if (v < 0) {
+                PyErr_SetString(PyExc_OverflowError, "negative value not allowed for uint64 property");
+                return -1;
+            }
+            g_value_set_uint64(value, v);
+        } else
+#endif
+        if (PyLong_Check(obj))
             g_value_set_uint64(value, PyLong_AsUnsignedLongLong(obj));
         else
             return -1;
@@ -788,19 +828,9 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	break;
     case G_TYPE_FLOAT:
 	g_value_set_float(value, PyFloat_AsDouble(obj));
-	if (PyErr_Occurred()) {
-	    g_value_unset(value);
-	    PyErr_Clear();
-	    return -1;
-	}
 	break;
     case G_TYPE_DOUBLE:
 	g_value_set_double(value, PyFloat_AsDouble(obj));
-	if (PyErr_Occurred()) {
-	    g_value_unset(value);
-	    PyErr_Clear();
-	    return -1;
-	}
 	break;
     case G_TYPE_STRING:
 	if (obj == Py_None)
@@ -895,6 +925,11 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	    break;
 	}
     }
+    if (PyErr_Occurred()) {
+        g_value_unset(value);
+        PyErr_Clear();
+        return -1;
+    }
     return 0;
 }
 
